
/* sigma compare, author Gary Demos June 2014 */
/* no warranties expressed nor implied */
/* no representation is herein made as to usefulness nor suitability for any pupose */
/* caution, code may contain bugs and design flaws */
/* use at your own risk */

// to build:
/*
 
 g++ sigma_compare.cpp dpx_file_io.cpp -o sigma_compare_linux -lpthread \
 -I /usr/local/include/OpenEXR \
 /usr/local/lib/libHalf.a /usr/local/lib/libIlmImf.a /usr/local/lib/libIex.a /usr/local/lib/libIlmThread.a \
 /usr/local/lib/libz.a -m64 -O1
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <math.h>

//#include <OpenEXR/ImfRgbaFile.h> /* OpenExr */
//#include <OpenEXR/ImfArray.h> /* OpenExr */
#include "/usr/local/include/OpenEXR/ImfRgbaFile.h" /* OpenExr */
#include "/usr/local/include/OpenEXR/ImfArray.h" /* OpenExr */
using namespace Imf; /* OpenExr */
using namespace Imath; /* OpenExr */

#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))

#define NUMBER_OF_STOPS 32 // note: the 5 bit half-float exponent is limited to a range of 31 stops (value 31 = +-inf nan out-of-range code).  Note: value 0-exp denorm extends lower),  Use 32-bit floating point I/O (e.g. dpx32) for ranges greater than 31 (+denorm) stops

// Prototypes:
void dpx_read (char *inname, float **pixels_read, short *width, short *height, short cineon, short half_flag);


#define SIGMA_MULTIPLES 7 /*2,3,4,6,8,12,16*/

typedef struct
{
    // measurements
    double square_sum[NUMBER_OF_STOPS];
    double sum[NUMBER_OF_STOPS];
    double count[NUMBER_OF_STOPS];   /* must be double precision 64-bit floating point to continue to accumuate counts */

    double count_neg;
    double sum_neg;
    double square_sum_neg;

    double count_multiple_sigma[SIGMA_MULTIPLES][NUMBER_OF_STOPS];
    double furthest_outlier_sum[NUMBER_OF_STOPS];
    
    
    // derived stats
    double average_neg;
    float sigma_neg;
    double MSE_neg;
    double self_relative_neg;
    double average[NUMBER_OF_STOPS];
    double self_relative[NUMBER_OF_STOPS];
    double MSE[NUMBER_OF_STOPS];
    float sigma[NUMBER_OF_STOPS];
    double furthest_outlier_avg[NUMBER_OF_STOPS];
    
    
} sigma_t;


void init_sigma_stats( sigma_t *sig )
{
    for (int j=0; j<NUMBER_OF_STOPS; j++) {
        sig->square_sum[j] = 0.0;
        sig->sum[j] = 0.0;
        sig->MSE[j] = 0.0;
        sig->average[j] = 0.0;
        sig->count[j] = 0;
        sig->furthest_outlier_sum[j] = 0;
        for(int k=0; k < SIGMA_MULTIPLES; k++) {
            sig->count_multiple_sigma[k][j] = 0;
        } /* k */
    } /* j */
    
    sig->square_sum_neg = 0.0;
    sig->sum_neg = 0.0;
    sig->count_neg = 0.0;
    sig->average_neg = 0.0;
    sig->MSE_neg = 0.0;
    
}

typedef struct {
    float *buf[3];
    int width;
    int height;
    int init;
} pic_t;


void init_pic( pic_t *pic, int width, int height )
{
    pic->width = width;
    pic->height = height;

    for(int c=0;c<3;c++)
        if( pic->buf[c] == NULL )
            pic->buf[c] = (float *) malloc( pic->width * pic->height * sizeof( float ));
    
    pic->init = 1;
}

void copy_exr_to_pic( pic_t *dst,   Array2D<Rgba> *half_float_pixels, float amplitude_factor )
{
    for(int y=0; y< dst->height; y++) {
        for(int x=0; x <  dst->width; x++) {
            dst->buf[0][ y * dst->width + x] = amplitude_factor * half_float_pixels[y][x]->r;
            dst->buf[1][ y * dst->width + x] = amplitude_factor * half_float_pixels[y][x]->g;
            dst->buf[2][ y * dst->width + x] = amplitude_factor * half_float_pixels[y][x]->b;
        } /* x */
    } /* y */
}


void copy_dpx_to_pic( pic_t *dst,  float *src, float amplitude_factor )
{
    for(int c = 0; c < 3; c ++ )
    {
        for(int y=0; y< dst->height; y++) {
            for(int x=0; x <  dst->width; x++) {
                dst->buf[c][ y * dst->width + x] = amplitude_factor * src[(c*dst->height + y) * dst->width + x];
            } /* x */
        } /* y */
    }
    
}

int bin_val( float sample )
{
    int j;
    
    if ( sample  < 0.0) {
        j = -1;
    } else { /* col[c] >= 0 */
        
        float flt_log2 = log2f(MAX(1e-16, sample));
   
        j = flt_log2; /* float to int */
    
        if (flt_log2 > 0.0) j = j + 1; /* must round one side of zero for continuity of integer j */
        
        j = MIN(NUMBER_OF_STOPS - 1, MAX(0, j + NUMBER_OF_STOPS / 2)); /* split NUMBER_OF_STOPS half above 1.0 and half below 1.0 */
    } /* col[c] < 0 or not */

    return(j);
}

void measure_frame_sigma_pass1( sigma_t *sig,  float *src_pic, float *test_pic, int width, int height )
{
    for(int y=0; y< height; y++)
    {
        for(int x=0; x< width; x++)
        {
            int j;
        
            float val = src_pic[ y * width + x];
        
            /* store difference from original in pixels2 */
            float dif = val - test_pic[ y * width + x];
        
            if (val < 0.0) {
                sig->square_sum_neg += dif * dif;
                sig->sum_neg += val;
                sig->count_neg= sig->count_neg + 1;
            } else { /* col[c] >= 0 */
                float flt_log2 = log2f(MAX(1e-16, val));
                j = flt_log2; /* float to int */
                if (flt_log2 > 0.0) j = j + 1; /* must round one side of zero for continuity of integer j */
                
                j = MIN(NUMBER_OF_STOPS - 1, MAX(0, j + NUMBER_OF_STOPS / 2)); /* split NUMBER_OF_STOPS half above 1.0 and half below 1.0 */
                sig->square_sum[j] += dif * dif;
                sig->sum[j] += val;
                sig->count[j]++;
            } /* col[c] < 0 or not */
        } /* x */
    } /* y */

}

/* turns k=0,1,2,3,4,5,6,7 into multiple=2,3,4,6,8,12,16 */

int k_multiple( int k )
{
    int multiple = (1 << ((k>>1) + 1));
    multiple = multiple + (k & 1) * (multiple >> 1);
    
    return (multiple);
}

void measure_frame_sigma_pass2( sigma_t *sig,  float *src_pic, float *test_pic, int width, int height )
{
    for(int y=0; y< height; y++)
    {
        for(int x=0; x< width; x++)
        {
            int j = bin_val( src_pic[ y* width + x ]  );
        
            if (j  >= 0)
            { /* dont count negs */
                float dif;
            
                if (sig->count[j] > 0)
                {
                    dif = src_pic[ y* width + x ] -  test_pic[y * width + x];
                
                    for(int k=0; k < SIGMA_MULTIPLES; k++)
                    {
                        short multiple = k_multiple(k);
                        
                        if (fabsf(dif) > (multiple * sig->sigma[j]))
                        {
                            sig->count_multiple_sigma[k][j]++;
                        
                            if(k== (SIGMA_MULTIPLES-1)) {
                            //  printf(" c=%d j=%d x=%d y=%d col_dif[%d]=%e\n", c, j, x, y, c, col_dif[c]);
                                sig->furthest_outlier_sum[j] += fabsf(dif);
                            } /* k== (SIGMA_MULTIPLES-1) */
                        } /* fabsf(col_dif[c]) > (multiple * sigma[j]) */
                    } /* k */
                } /* count[j] > 0 */
            } /* bin_value[(c*v_reso + y) * h_reso + x] >= 0 */
        
        } /* x */
    } /* y */
}




void read_one_frame( pic_t *dst_pic, char *filename, int dpx_in_file, short *h_reso, short *v_reso, int amplitude_factor )
{
    
    if (dpx_in_file == 1)
    {
        float *pixels = NULL;
    
        dpx_read (filename, &pixels, h_reso, v_reso, 0/*not cineon*/, 0/*no half_flag*/);
    
        init_pic( dst_pic, *h_reso, *v_reso );
        copy_dpx_to_pic( dst_pic,  pixels,  amplitude_factor );
    
    
        free(pixels);
    
    } else { /* exr */
    
        static Array2D<Rgba> half_float_pixels; /* unused if both input files are dpx */
        
        RgbaInputFile file (filename, 1);
        
        Box2i dw = file.dataWindow();
        
        *h_reso = dw.max.x - dw.min.x + 1;
        *v_reso = dw.max.y - dw.min.y + 1;
        half_float_pixels.resizeErase (*v_reso, *h_reso);
        
        file.setFrameBuffer (&half_float_pixels[0][0] - dw.min.x - dw.min.y * *h_reso, 1, *h_reso);
        file.readPixels (dw.min.y, dw.max.y);
        
        printf(" reading file %s having h_reso = %d v_reso = %d\n", filename, *h_reso, *v_reso);
        
        init_pic( dst_pic, *h_reso, *v_reso );
        
        for(int y=0; y< dst_pic->height; y++) {
            for(int x=0; x <  dst_pic->width; x++) {
                dst_pic->buf[0][ y * dst_pic->width + x] = amplitude_factor * half_float_pixels[y][x].r;
                dst_pic->buf[1][ y * dst_pic->width + x] = amplitude_factor * half_float_pixels[y][x].g;
                dst_pic->buf[2][ y * dst_pic->width + x] = amplitude_factor * half_float_pixels[y][x].b;
            } /* x */
        } /* y */
    
    } /* dpx vs exr */
}

const char *c_name[] = {"red","grn","blu"};

// calculations separate from print for iteration purposes (as in machine learning, coefficient design)
void calc_pass1_stats( sigma_t *sigma_stats_array, float amplitude_factor )
{
    for (int c=0; c<3; c++)
    {
        sigma_t *sig = &(sigma_stats_array[c]);
        
        if (sig->count_neg > 0)
        {
            sig->MSE_neg  = sig->square_sum_neg / sig->count_neg;
            sig->sigma_neg = sqrtf(sig->MSE_neg);
            sig->average_neg = sig->sum_neg / sig->count_neg;
            sig->self_relative_neg = sig->sigma_neg / sig->average_neg;
        } /* count_neg[c] > 0 */
        
        for (int j=0; j<NUMBER_OF_STOPS; j++)
        {
            if (sig->count[j] > 0)
            {
                sig->MSE[j] = sig->square_sum[j] / sig->count[j];
                sig->sigma[j] = sqrtf(sig->MSE[j]);
                sig->average[j] = sig->sum[j]  / sig->count[j];
                sig->self_relative[j] = sig->sigma[j] / sig->average[j];
            } /* count[j] > 0 */
        } /* j */
    } /* c */
}

void print_pass1_stats( sigma_t *sigma_stats_array, float amplitude_factor )
{
    printf("\n amplitude_factor = %f\n", amplitude_factor);

    for (int c=0; c<3; c++)
    {
        sigma_t *sig = &(sigma_stats_array[c]);
    
        printf("\n");
        if (sig->count_neg > 0)
        {
            printf(" sigma_neg_%s = %e self_relatve = %f (%f%%) at average value = %e for %.0f pixels\n", c_name[c],
                    sig->sigma_neg,
                    sig->self_relative_neg,
                    100.0 * sig->self_relative_neg,
                    sig->average_neg,
                    sig->count_neg);
        
        } /* count_neg[c] > 0 */
    
        for (int j=0; j<NUMBER_OF_STOPS; j++)
        {
            if (sig->count[j] > 0)
            {
                printf(" sigma_%s[%d] = %e self_relative = %f (%f%%) at average value = %e for %.0f pixels\n", c_name[c],
                               j, sig->sigma[j], sig->self_relative[j], 100.0 * sig->self_relative[j], sig->average[j], sig->count[j]);
            
            } /* count[j] > 0 */
        } /* j */
    } /* c */
}



void print_pass2_stats( sigma_t *sigma_stats_array, float amplitude_factor )
{

    
    
    short multiple;
    for (int c=0; c<3; c++) {
        sigma_t *sig = &(sigma_stats_array[c]);
        
        printf("\n");
        for (int j=0; j<NUMBER_OF_STOPS; j++) {
            if ((sig->count[j] > 0) && (sig->count_multiple_sigma[0][j] > 0)) {
                printf("  sigma_%s[%d] = %e self_relative = %f (%f%%) at average value = %e for %.0f pixels\n", c_name[c],
                                   j, sig->sigma[j], sig->self_relative[j], 100.0 * sig->self_relative[j], sig->average[j], sig->count[j]);
            } /* count[j] > 0 */
            for( int k=0; k < (SIGMA_MULTIPLES-1); k++)
            {
                if ((sig->count_multiple_sigma[k][j] > 0) && (sig->count_multiple_sigma[k][j] > sig->count_multiple_sigma[k+1][j]))
                {
                    multiple = k_multiple( k );
                    
                    printf(" %dsigma_%s[%d] = %e self_relative = %f (%f%%) at average value = %e for %.0f pixels which is %f %% of the %.0f pixels within a stop of this value\n",
                        multiple, c_name[c], j, multiple*sig->sigma[j], multiple*sig->self_relative[j], 100.0 * multiple*sig->self_relative[j], sig->average[j], sig->count_multiple_sigma[k][j], (sig->count_multiple_sigma[k][j] * 100.0)/(1.0 * sig->count[j]), sig->count[j]);
                } /* count_multiple_sigma[k][j] > 0 */
            } /* k */
        
            int k = SIGMA_MULTIPLES-1;
        
            if (sig->count_multiple_sigma[k][j] > 0)
            {
                sig->furthest_outlier_avg[j] = sig->furthest_outlier_sum[j] / sig->count_multiple_sigma[k][j]; /* compute average from sum */
                multiple = (1 << ((k>>1) + 1));
                multiple = multiple + (k & 1) * (multiple >> 1); /* turns k=0,1,2,3,4,5,6,7 into multiple=2,3,4,6,8,12,16 */
                printf(" beyond %dsigma_%s[%d] furthest_outlier_avg dif = %e which is %f sigma (sigma being %e), self_relative = %f (%f%%) at average value = %e for %.0f pixels which is %f %% of the %.0f pixels within a stop of this value\n",
                                   multiple, c_name[c], j, sig->furthest_outlier_avg[j], sig->furthest_outlier_avg[j] / sig->sigma[j], sig->sigma[j], sig->furthest_outlier_avg[j] / sig->average[j], 100.0 * sig->furthest_outlier_avg[j] / sig->average[j], sig->average[j], sig->count_multiple_sigma[k][j], (sig->count_multiple_sigma[k][j] * 100.0)/(1.0 * sig->count[j]), sig->count[j]);
            } /* (count_multiple_sigma[k][j] > 0) */
        } /* j */
    } /* c */
    
}



/***********************************************************************************************************/
int
main(int argc, char **argv)
{
    int first_frame=0, last_frame=0;
    
    short h_reso, v_reso;

    sigma_t sigma_stats[3];

    short dpx_in_file1=0, dpx_in_file2=0;
    float amplitude_factor;
    
    pic_t src_pic, test_pic;
    
    for(int c=0;c<3;c++ )
        src_pic.buf[c] = test_pic.buf[c] = NULL;

    if (argc <= 2) {
        printf(" usage: %s original_files, test_comparison_files, first_frame, last_frame, amplitude_factor\n", argv[0]);
        printf(" note: use %% format for frame numbers, such as %%07d, within the file names\n");
        exit(1);
    }
    
    if (argc > 4) {
        first_frame = atoi(argv[3]);
        last_frame = atoi(argv[4]);
        printf(" first_frame = %d last_frame = %d\n", first_frame, last_frame);
        
    } else { /* argc <= 4 */
        printf(" no frame range, one frame will be processed\n");
    } /* argc > 4 or not */
    
    if (argc > 5) {
        amplitude_factor = atof(argv[5]);
        printf(" setting amplitude_factor to %f\n", amplitude_factor);
        /* can scale nits to unity relationship here
         e.g. if 1.0=100nits, use 100.0 here for the amplitude_factor in order to see the printout relative to nits (sigma and average value), does not affect self_relative
         e.g. also if looking at nits, but wish to see in terms of 100nits = 1.0, then use .01 for amplitude_factor to range 100nits down to 1.0 (e.g. a nominal white at 1.0) */
    } else { /* argc <= 5 */
        amplitude_factor = 1.0;
        printf(" using default amplitude_factor = %f\n", amplitude_factor);
    } /* argc > 5 or not */
    
    
    for(int c=0;c<3;c++)
        init_sigma_stats( &(sigma_stats[c]) );
    
    
    
    // PASS 1
    
    for (int iframe = first_frame; iframe <= last_frame; iframe++) {
        char infile1[300], infile2[300];

        sprintf(infile1, argv[1], iframe);
        sprintf(infile2, argv[2], iframe);
        
        char *path1 = argv[1];
        char *path2 = argv[2];
        
        short num_chars = strlen(infile1); /* length of in_file_name string */
        
        if ((!strcmp(&path1[num_chars-1],  "x"))  ||(!strcmp(&path1[num_chars-1],    "X")) || /* DPX file ending in ".dpx" (not recommended, 10bit linear has insufficient precision near black) */
            (!strcmp(&path1[num_chars-3],  "x32"))||(!strcmp(&path1[num_chars-3],  "X32")) || /* dpx32 or DPX32 */
            (!strcmp(&path1[num_chars-3],  "x16"))||(!strcmp(&path1[num_chars-3],  "X16")) || /* dpx16 or DPX16 (not recommended) */
            (!strcmp(&path1[num_chars-4], "xhlf"))||(!strcmp(&path1[num_chars-4], "XHLF"))) { /* dpxhlf or DPXHLF, 16-bit half-float (non-standard, improperly defined in dpx 2.0 standard) */
            dpx_in_file1 = 1;
        }
        num_chars = strlen(infile2); /* length of in_file_name string */
        
        if ((!strcmp(&path2[num_chars-1],  "x"))  ||(!strcmp(&infile2[num_chars-1],    "X")) || /* DPX file ending in ".dpx" (not recommended, 10bit linear has insufficient precision near black) */
            (!strcmp(&path2[num_chars-3],  "x32"))||(!strcmp(&path2[num_chars-3],  "X32")) || /* dpx32 or DPX32 */
            (!strcmp(&path2[num_chars-3],  "x16"))||(!strcmp(&path2[num_chars-3],  "X16")) || /* dpx16 or DPX16 (not recommended) */
            (!strcmp(&path2[num_chars-4], "xhlf"))||(!strcmp(&path2[num_chars-4], "XHLF"))) { /* dpxhlf or DPXHLF, 16-bit half-float (non-standard, improperly defined in dpx 2.0 standard) */
            dpx_in_file2 = 1;
        }
        
        read_one_frame( &test_pic, infile2, dpx_in_file2, &h_reso, &v_reso,  amplitude_factor );
        read_one_frame( &src_pic, infile1, dpx_in_file1, &h_reso, &v_reso,  amplitude_factor );

        
        for (int c=0; c<3; c++) {
            sigma_t *sig = &(sigma_stats[c]);
            
            measure_frame_sigma_pass1( sig,  src_pic.buf[c], test_pic.buf[c], h_reso, v_reso );
        } /* c */
        
    
        
    } /* iframe */
    
 
    
    calc_pass1_stats( &(sigma_stats[0]),  amplitude_factor );
    print_pass1_stats( &(sigma_stats[0]),  amplitude_factor );

    
    // PASS 2
    
    printf("\n beginning second pass over frames (multiples of sigma, now that sigma for all frames has been determined)\n\n");

    for (int iframe = first_frame; iframe <= last_frame; iframe++) {
        char infile1[300], infile2[300];

        sprintf(infile1, argv[1], iframe);
        sprintf(infile2, argv[2], iframe);
        
        read_one_frame( &test_pic, infile2, dpx_in_file2, &h_reso, &v_reso,  amplitude_factor );
        read_one_frame( &src_pic, infile1, dpx_in_file1, &h_reso, &v_reso,  amplitude_factor );

        for (int c=0; c<3; c++) {
            measure_frame_sigma_pass2( &(sigma_stats[c]),  src_pic.buf[c], test_pic.buf[c], h_reso, v_reso );
            
        } /* c */
       

        
    } /* iframe */
    
    print_pass1_stats( &(sigma_stats[0]),  amplitude_factor );  // repeated for some reason in Gary's original

    // calc. and print self relative
    print_pass2_stats( &(sigma_stats[0]),  amplitude_factor );
    
} /* main */



